/*****************************************************************

  Numerical Python interface to HDF-EOS

  Written by Dominic Mazzoni
  Copyright 2003
  NASA Jet Propulsion Laboratory
  California Institute of Technology

*****************************************************************/

#include <Python.h>
#include <Numeric/arrayobject.h>

#include "hdf.h"
#include "mfhdf.h"
#include "HdfEosDef.h"

#define MAX_FILES 16
#define MAX_GRIDS 64
#define MAX_FIELDS 1024
#define MAX_FIELD_NAME_LEN 64
#define MAX_DIMS 8

enum hdfeos_apis {
   grid_api,
   swath_api,
   point_api
};

typedef struct {
   int api;
   int32 id;
   char *name;
} hdfeos_open_file;

hdfeos_open_file g_files[MAX_FILES];
int g_num_files = 0;

/* These two functions have to be declared unless we link to the
   GCTP library (for coordinate transformation / projection
*/

void for_init(int32 projcode,
			  int32 zonecode,
			  float64 projparm[],
			  int32 spherecode,
			  void *z1,
			  void *z2,
			  int32 *errorcode,
			  int32 (*for_trans[100]) ())
{
  *errorcode = -1;
}

void inv_init(int32 projcode,
			  int32 zonecode,
			  float64 projparm[],
			  int32 spherecode,
			  void *z1,
			  void *z2,
			  int32 *errorcode,
			  int32 (*inv_trans[100]) ())
{
  *errorcode = -1;
}

/**************
 * Utility fns
 **************/

static void close_hdf_file(hdfeos_open_file *f)
{
   switch(f->api) {
   case grid_api:
      GDclose(f->id);
      break;
   case swath_api:
      SWclose(f->id);
      break;
   case point_api:
      PTclose(f->id);
      break;
   }
   free(f->name);
}

static int32 open_hdf_file(char *filename, int api)
{
   int i, j;
   int32 id;

   for(i=0; i<g_num_files; i++) {
      if (api == g_files[i].api &&
          !strcmp(filename, g_files[i].name)) {
         /* move to front of list so that list is in
            most-recently-used order */
         hdfeos_open_file tmp = g_files[i];
         for(j=i; j>0; j--)
            g_files[j] = g_files[j-1];
         g_files[0] = tmp;
         return g_files[0].id;
      }
   }

   /* It wasn't in the list, so we need to open a new file. */

   /* First make sure we don't have too many files open (HDF has
      a limit); if so, close the least-recently-used one. */
   if (g_num_files == MAX_FILES) {
      /* Always close the last file, since every access to a file
         moves it to the front of the list */
      i = g_num_files-1;
      close_hdf_file(&g_files[i]);
      g_num_files--;
   }

   /* Try to open the file */

   id = -1;
   switch(api) {
   case grid_api:
      id = GDopen(filename, DFACC_READ);
      break;
   case swath_api:
      id = SWopen(filename, DFACC_READ);
      break;
   case point_api:
      id = PTopen(filename, DFACC_READ);
      break;
   }

   if (id < 0)
      return id;

   /* If it succeeded, insert it in the front of the list */
   for(i=g_num_files; i>0; i--)
      g_files[i] = g_files[i-1];
   g_num_files++;
   g_files[0].api = api;
   g_files[0].name = (char *)malloc(strlen(filename)+1);
   strcpy(g_files[0].name, filename);
   g_files[0].id = id;

   return id;
}

static PyObject *error_str(PyObject *exception, char *str)
{
   PyErr_SetString(PyExc_IOError, str);
   return NULL;
}

static PyObject *error_str_s(PyObject *exception, char *format, char *arg)
{
   char *str = malloc(strlen(arg)+strlen(format)+50);
   sprintf(str, format, arg);
   PyErr_SetString(PyExc_IOError, str);
   free(str);
   return NULL;
}

static PyObject *error_str_d(PyObject *exception, char *format, int arg)
{
   char *str = malloc(strlen(format)+50);
   sprintf(str, format, arg);
   PyErr_SetString(PyExc_IOError, str);
   free(str);
   return NULL;
}

/* Converts comma-separated-values to a Python tuple */
PyObject *csv_to_tuple(int len, char *str)
{
   PyObject *result = PyTuple_New(len);
   int i;

   for(i=0; i<len; i++) {
      if (strstr(str, ","))
         strstr(str, ",")[0] = 0;
      PyTuple_SetItem(result, i, Py_BuildValue("s", str));
      str += strlen(str)+1;
   }

   return result;
}

int HDF_to_numpy_type(int32 numbertype)
{
   switch(numbertype) {
   case DFNT_INT8:
      return PyArray_SBYTE;
   case DFNT_UINT8:
      return PyArray_UBYTE;
   case DFNT_INT16:
      return PyArray_SHORT;
   case DFNT_UINT16:
      return PyArray_USHORT;
   case DFNT_INT32:
      return PyArray_INT;
   case DFNT_UINT32:
      return PyArray_UINT;
   case DFNT_FLOAT32:
      return PyArray_FLOAT;
   case DFNT_FLOAT64:
      return PyArray_DOUBLE;
   default:
      return 0;
   }
}

static PyObject *
hdfeos_close_all(PyObject *self, PyObject *args)
{
   int i;
   
   for(i=0; i<g_num_files; i++)
      close_hdf_file(&g_files[i]);
   g_num_files = 0;

   Py_INCREF(Py_None);
   return Py_None;
}

/**************
 * Grid API
 **************/

static PyObject *open_grid_dataset(char *filename, char *dataset,
                                   int32 *gdfid, int32 *GDid)
{
   *gdfid = open_hdf_file(filename, grid_api);
   if (*gdfid < 0)
      return error_str_s(PyExc_IOError,
                         "Could not open grid interface to %s\n", filename);

   *GDid = GDattach(*gdfid, dataset);
   if (*GDid < 0)
      return error_str_s(PyExc_IOError,
                         "Could not attach to grid dataset %s\n", dataset);

   return NULL;
}

static PyObject *
hdfeos_grids(PyObject *self, PyObject *args)
{
   PyObject *result;
   char *filename;
   char *gridlist;
   int32 bufferlen;
   int num_grids;

   if (!PyArg_ParseTuple(args, "s", &filename))
      return NULL;

   num_grids = GDinqgrid(filename, NULL, &bufferlen);
   if (num_grids <= 0)
      return error_str_s(PyExc_IOError,
                         "No grids found in %s\n", filename);

   gridlist = (char *)malloc(bufferlen+1);
   num_grids = GDinqgrid(filename, gridlist, &bufferlen);

   result = csv_to_tuple(num_grids, gridlist);
   free(gridlist);

   return result;
}

static PyObject *
hdfeos_grid_fields(PyObject *self, PyObject *args)
{
   PyObject *result;
   char *filename;
   char *dataset;
   char *fieldlist;
   int num_fields;
   int32 *rank;
   int32 *numbertype;
   int32 gdfid;
   int32 GDid;

   if (!PyArg_ParseTuple(args, "ss", &filename, &dataset))
      return NULL;

   result = open_grid_dataset(filename, dataset, &gdfid, &GDid);
   if (result)
      return result;

   fieldlist = (char *)malloc(MAX_FIELDS * MAX_FIELD_NAME_LEN);
   rank = (int32 *)malloc(MAX_FIELDS * sizeof(int32));
   numbertype = (int32 *)malloc(MAX_FIELDS * sizeof(int32));
   num_fields = GDinqfields(GDid, fieldlist, rank, numbertype);

   result = csv_to_tuple(num_fields, fieldlist);
   free(fieldlist);
   free(rank);
   free(numbertype);

   GDdetach(GDid);

   return result;
}

static PyObject *
hdfeos_grid_field_dims(PyObject *self, PyObject *args)
{
   PyObject *result;
   char *filename, *dataset, *field;
   int32 gdfid;
   int32 GDid;
   int32 rank;
   int32 dims[MAX_DIMS];
   int32 numbertype;
   intn  status;
   int   i;
   char  dimlist[1000];

   if (!PyArg_ParseTuple(args, "sss", &filename, &dataset, &field))
      return NULL;

   result = open_grid_dataset(filename, dataset, &gdfid, &GDid);
   if (result)
      return result;

   status = GDfieldinfo(GDid, field, &rank, dims, &numbertype, dimlist);
   if (status < 0) {
      GDdetach(GDid);
      return error_str_s(PyExc_IOError,
                         "Could not read from field %s\n", field);
   }

   result = PyTuple_New(rank);
   for(i=0; i<rank; i++)
      PyTuple_SetItem(result, i, Py_BuildValue("i", (int)dims[i]));

   GDdetach(GDid);

   return result;
}

static PyObject *
hdfeos_grid_field_read(PyObject *self, PyObject *args)
{
   int numpy_type;
   PyObject *result;
   PyObject *range[MAX_DIMS];
   char *filename, *dataset, *field;
   int32 gdfid;
   int32 GDid;
   int32 rank;
   int32 dims[MAX_DIMS];
   int   pyrank;
   int   pydims[MAX_DIMS];
   int32 start[MAX_DIMS], stride[MAX_DIMS], edge[MAX_DIMS];
   int32 numbertype;
   intn  status;
   int   in_ranges;
   int   first_range;
   int   i;
   char  dimlist[1000];

   if (!PyArg_ParseTuple(args, "sss|OOOOOOOO",
                         &filename, &dataset, &field,
                         &range[0], &range[1], &range[2], &range[3],
                         &range[4], &range[5], &range[6], &range[7]))
      return NULL;

   first_range = 3;
   in_ranges = PyTuple_Size(args) - first_range;
   
   result = open_grid_dataset(filename, dataset, &gdfid, &GDid);
   if (result)
      return result;

   status = GDfieldinfo(GDid, field, &rank, dims, &numbertype, dimlist);
   if (status < 0) {
      GDdetach(GDid);
      return error_str_s(PyExc_IOError,
                         "Could not read from field %s\n", field);
   }

   numpy_type = HDF_to_numpy_type(numbertype);
   if (numpy_type == 0) {
      GDdetach(GDid);
      return error_str_d(PyExc_IOError,
                         "Unsupported HDF numeric type: %d", numbertype);
   }

   for(i=0; i<rank; i++) {
      if (i < in_ranges) {
         if (PyTuple_Check(range[i])) {
            if (PyTuple_Size(range[i]) == 2) {
               start[i] = PyInt_AsLong(PyTuple_GetItem(range[i], 0));
               edge[i] = PyInt_AsLong(PyTuple_GetItem(range[i], 1))
                  - start[i];
               stride[i] = 1;
            }
            else if (PyTuple_Size(range[i]) == 3) {
               start[i] = PyInt_AsLong(PyTuple_GetItem(range[i], 0));
               stride[i] = PyInt_AsLong(PyTuple_GetItem(range[i], 1));
               edge[i] = (PyInt_AsLong(PyTuple_GetItem(range[i], 2))
                          - start[i]) / stride[i];
            }
            else {
               GDdetach(GDid);
               return error_str(PyExc_TypeError,
                                "Range tuples must contain exactly "
                                "2 or 3 elements.");
            }
         }
         else if (PyInt_Check(range[i])) {
            start[i] = (int32)PyInt_AsLong(range[i]);
            stride[i] = 1;
            edge[i] = 1;
         }
         else {
            GDdetach(GDid);
            return error_str(PyExc_TypeError,
                             "Range parameters must be either an "
                             "integer, a (start, end) tuple, or "
                             "a (start, stride, end) tuple.");
         }
      }
      else {
         start[i] = 0;
         stride[i] = 1;
         edge[i] = dims[i];
      }
   }

   for(i=0; i<rank; i++) {
      if (start[i] < 0 || start[i] >= dims[i] ||
          edge[i] < 0 || edge[i] > dims[i]) {
         GDdetach(GDid);
         return error_str(PyExc_ValueError,
                          "Range parameters are not within bounds.");
      }

      pydims[i] = edge[i];
   }

   pyrank = rank;

   while(pyrank > 0 && pydims[0] == 1) {
      for(i=0; i<8; i++)
         pydims[i] = pydims[i+1];
      pyrank--;
   }

   result = PyArray_FromDims(pyrank, pydims, numpy_type);
   
   status = GDreadfield(GDid, field, start, stride, edge,
                        ((PyArrayObject *)result)->data);

   if (status < 0) {
      Py_DECREF(result);
      GDdetach(GDid);
      return error_str_s(PyExc_IOError,
                         "Could not read from field %s\n", field);
   }

   GDdetach(GDid);

   return result;
}

/**************
 * Swath API
 **************/

static PyObject *open_swath_dataset(char *filename, char *dataset,
                                   int32 *swfid, int32 *SWid)
{
   *swfid = open_hdf_file(filename, swath_api);
   if (*swfid < 0)
      return error_str_s(PyExc_IOError,
                         "Could not open swath interface to %s\n", filename);

   *SWid = SWattach(*swfid, dataset);
   if (*SWid < 0)
      return error_str_s(PyExc_IOError,
                         "Could not attach to swath dataset %s\n", dataset);

   return NULL;
}

static PyObject *
hdfeos_swaths(PyObject *self, PyObject *args)
{
   PyObject *result;
   char *filename;
   char *swathlist;
   int32 bufferlen;
   int num_swaths;

   if (!PyArg_ParseTuple(args, "s", &filename))
      return NULL;

   num_swaths = SWinqswath(filename, NULL, &bufferlen);
   if (num_swaths <= 0)
      return error_str_s(PyExc_IOError,
                         "No swaths found in %s\n", filename);

   swathlist = (char *)malloc(bufferlen+1);
   num_swaths = SWinqswath(filename, swathlist, &bufferlen);

   result = csv_to_tuple(num_swaths, swathlist);
   free(swathlist);

   return result;
}

static PyObject *
hdfeos_swath_data_fields(PyObject *self, PyObject *args)
{
   PyObject *result;
   char *filename;
   char *dataset;
   char *fieldlist;
   int num_fields;
   int32 *rank;
   int32 *numbertype;
   int32 swfid;
   int32 SWid;

   if (!PyArg_ParseTuple(args, "ss", &filename, &dataset))
      return NULL;

   result = open_swath_dataset(filename, dataset, &swfid, &SWid);
   if (result)
      return result;

   fieldlist = (char *)malloc(MAX_FIELDS * MAX_FIELD_NAME_LEN);
   rank = (int32 *)malloc(MAX_FIELDS * sizeof(int32));
   numbertype = (int32 *)malloc(MAX_FIELDS * sizeof(int32));
   num_fields = SWinqdatafields(SWid, fieldlist, rank, numbertype);

   result = csv_to_tuple(num_fields, fieldlist);
   free(fieldlist);
   free(rank);
   free(numbertype);

   SWdetach(SWid);

   return result;
}

static PyObject *
hdfeos_swath_geo_fields(PyObject *self, PyObject *args)
{
   PyObject *result;
   char *filename;
   char *dataset;
   char *fieldlist;
   int num_fields;
   int32 *rank;
   int32 *numbertype;
   int32 swfid;
   int32 SWid;

   if (!PyArg_ParseTuple(args, "ss", &filename, &dataset))
      return NULL;

   result = open_swath_dataset(filename, dataset, &swfid, &SWid);
   if (result)
      return result;

   fieldlist = (char *)malloc(MAX_FIELDS * MAX_FIELD_NAME_LEN);
   rank = (int32 *)malloc(MAX_FIELDS * sizeof(int32));
   numbertype = (int32 *)malloc(MAX_FIELDS * sizeof(int32));
   num_fields = SWinqgeofields(SWid, fieldlist, rank, numbertype);

   result = csv_to_tuple(num_fields, fieldlist);
   free(fieldlist);
   free(rank);
   free(numbertype);

   SWdetach(SWid);

   return result;
}

static PyObject *
hdfeos_swath_field_dims(PyObject *self, PyObject *args)
{
   PyObject *result;
   char *filename, *dataset, *field;
   int32 swfid;
   int32 SWid;
   int32 rank;
   int32 dims[MAX_DIMS];
   int32 numbertype;
   intn  status;
   int   i;
   char  dimlist[1000];

   if (!PyArg_ParseTuple(args, "sss", &filename, &dataset, &field))
      return NULL;

   result = open_swath_dataset(filename, dataset, &swfid, &SWid);
   if (result)
      return result;

   status = SWfieldinfo(SWid, field, &rank, dims, &numbertype, dimlist);
   if (status < 0) {
      SWdetach(SWid);
      return error_str_s(PyExc_IOError,
                         "Could not read from field %s\n", field);
   }

   result = PyTuple_New(rank);
   for(i=0; i<rank; i++)
      PyTuple_SetItem(result, i, Py_BuildValue("i", (int)dims[i]));

   SWdetach(SWid);

   return result;
}

static PyObject *
hdfeos_swath_field_read(PyObject *self, PyObject *args)
{
   int numpy_type;
   PyObject *result;
   PyObject *range[MAX_DIMS];
   char *filename, *dataset, *field;
   int32 swfid;
   int32 SWid;
   int32 rank;
   int32 dims[MAX_DIMS];
   int   pyrank;
   int   pydims[MAX_DIMS];
   int32 start[MAX_DIMS], stride[MAX_DIMS], edge[MAX_DIMS];
   int32 numbertype;
   intn  status;
   int   in_ranges;
   int   first_range;
   int   i;
   char  dimlist[1000];

   if (!PyArg_ParseTuple(args, "sss|OOOOOOOO",
                         &filename, &dataset, &field,
                         &range[0], &range[1], &range[2], &range[3],
                         &range[4], &range[5], &range[6], &range[7]))
      return NULL;

   first_range = 3;
   in_ranges = PyTuple_Size(args) - first_range;
   
   result = open_swath_dataset(filename, dataset, &swfid, &SWid);
   if (result)
      return result;

   status = SWfieldinfo(SWid, field, &rank, dims, &numbertype, dimlist);
   if (status < 0) {
      SWdetach(SWid);
      return error_str_s(PyExc_IOError,
                         "Could not read from field %s\n", field);
   }

   numpy_type = HDF_to_numpy_type(numbertype);
   if (numpy_type == 0) {
      SWdetach(SWid);
      return error_str_d(PyExc_IOError,
                         "Unsupported HDF numeric type: %d", numbertype);
   }

   for(i=0; i<rank; i++) {
      if (i < in_ranges) {
         if (PyTuple_Check(range[i])) {
            if (PyTuple_Size(range[i]) == 2) {
               start[i] = PyInt_AsLong(PyTuple_GetItem(range[i], 0));
               edge[i] = PyInt_AsLong(PyTuple_GetItem(range[i], 1))
                  - start[i];
               stride[i] = 1;
            }
            else if (PyTuple_Size(range[i]) == 3) {
               start[i] = PyInt_AsLong(PyTuple_GetItem(range[i], 0));
               stride[i] = PyInt_AsLong(PyTuple_GetItem(range[i], 1));
               edge[i] = (PyInt_AsLong(PyTuple_GetItem(range[i], 2))
                          - start[i]) / stride[i];
            }
            else {
               SWdetach(SWid);
               return error_str(PyExc_TypeError,
                                "Range tuples must contain exactly "
                                "2 or 3 elements.");
            }
         }
         else if (PyInt_Check(range[i])) {
            start[i] = (int32)PyInt_AsLong(range[i]);
            stride[i] = 1;
            edge[i] = 1;
         }
         else {
            SWdetach(SWid);
            return error_str(PyExc_TypeError,
                             "Range parameters must be either an "
                             "integer, a (start, end) tuple, or "
                             "a (start, stride, end) tuple.");
         }
      }
      else {
         start[i] = 0;
         stride[i] = 1;
         edge[i] = dims[i];
      }
   }

   for(i=0; i<rank; i++) {
      if (start[i] < 0 || start[i] >= dims[i] ||
          edge[i] < 0 || edge[i] > dims[i]) {
         SWdetach(SWid);
         return error_str(PyExc_ValueError,
                          "Range parameters are not within bounds.");
      }

      pydims[i] = edge[i];
   }

   pyrank = rank;

   while(pyrank > 0 && pydims[0] == 1) {
      for(i=0; i<8; i++)
         pydims[i] = pydims[i+1];
      pyrank--;
   }

   result = PyArray_FromDims(pyrank, pydims, numpy_type);
   
   status = SWreadfield(SWid, field, start, stride, edge,
                        ((PyArrayObject *)result)->data);

   if (status < 0) {
      Py_DECREF(result);
      SWdetach(SWid);
      return error_str_s(PyExc_IOError,
                         "Could not read from field %s\n", field);
   }

   SWdetach(SWid);

   return result;
}

static PyMethodDef hdfeos_methods[] = {
   {"grids", hdfeos_grids, METH_VARARGS,
    "Retrieve list of all grids in the file."},
   {"grid_fields", hdfeos_grid_fields, METH_VARARGS,
    "Retrieve list of all fields in a grid."},
   {"grid_field_dims", hdfeos_grid_field_dims, METH_VARARGS,
    "Get the dimensions of a field from a Grid dataset."},
   {"grid_field_read", hdfeos_grid_field_read, METH_VARARGS,
    "Read data from a Grid field."},
   {"swaths", hdfeos_swaths, METH_VARARGS,
    "Retrieve list of all swaths in the file."},
   {"swath_data_fields", hdfeos_swath_data_fields, METH_VARARGS,
    "Retrieve list of all data fields in a swath."},
   {"swath_geo_fields", hdfeos_swath_geo_fields, METH_VARARGS,
    "Retrieve list of all geolocation fields in a swath."},
   {"swath_field_dims", hdfeos_swath_field_dims, METH_VARARGS,
    "Get the dimensions of a (data or geo) field from a Swath dataset."},
   {"swath_field_read", hdfeos_swath_field_read, METH_VARARGS,
    "Read data from a Swath field (data or geo)."},
   {"close_all", hdfeos_close_all, METH_VARARGS,
    "Closes all open HDF files."},
   {NULL, NULL, 0, NULL}        /* Sentinel */
};

PyMODINIT_FUNC
inithdfeos(void)
{
   PyObject *m;
   
   m = Py_InitModule("hdfeos", hdfeos_methods);
   import_array();
}
